@@ -9,56 +9,51 @@
 //! [`AlgoSimple`] offers a shorthand for specifying an algorithm,
 //! specifically for key generation on the card.
 
+use crate::card_do::ApplicationRelatedData;
 use crate::crypto_data::EccType;
-use crate::{Error, KeyType};
+use crate::{keys, Error, KeyType};
 
-use anyhow::anyhow;
+use anyhow::{anyhow, Result};
 use std::convert::TryFrom;
 use std::fmt;
 
 /// A shorthand way to specify algorithms (e.g. for key generation).
-///
-/// RSA variants require "number of bits in 'e'" as parameter.
-///
-/// There are (at least) two common supported values for e:
-///  e=17 [YK4, YK5]
-///  e=32 [YK5, Floss3.4, Gnuk1.2]
 #[derive(Clone, Copy, Debug)]
 #[non_exhaustive]
 pub enum AlgoSimple {
-    RSA1k(u16),
-    RSA2k(u16),
-    RSA3k(u16),
-    RSA4k(u16),
+    RSA1k,
+    RSA2k,
+    RSA3k,
+    RSA4k,
     NIST256,
     NIST384,
     NIST521,
     Curve25519,
 }
 
-impl From<&str> for AlgoSimple {
-    fn from(algo: &str) -> Self {
+impl TryFrom<&str> for AlgoSimple {
+    type Error = anyhow::Error;
+
+    fn try_from(algo: &str) -> Result<Self, anyhow::Error> {
         use AlgoSimple::*;
 
-        match algo {
-            "RSA2k/17" => RSA2k(17),
-            "RSA2k/32" => RSA2k(32),
-            "RSA3k/17" => RSA3k(17),
-            "RSA3k/32" => RSA3k(32),
-            "RSA4k/17" => RSA4k(17),
-            "RSA4k/32" => RSA4k(32),
+        Ok(match algo {
+            "RSA1k" => RSA1k,
+            "RSA2k" => RSA2k,
+            "RSA3k" => RSA3k,
+            "RSA4k" => RSA4k,
             "NIST256" => NIST256,
             "NIST384" => NIST384,
             "NIST521" => NIST521,
             "Curve25519" => Curve25519,
-            _ => panic!("unexpected algo {}", algo),
-        }
+            _ => return Err(anyhow!("unexpected algo {}", algo)),
+        })
     }
 }
 
 impl AlgoSimple {
     /// Get corresponding EccType by KeyType (except for Curve25519)
-    fn get_ecc_type(key_type: KeyType) -> EccType {
+    fn ecc_type(key_type: KeyType) -> EccType {
         match key_type {
             KeyType::Signing
             | KeyType::Authentication
@@ -68,7 +63,7 @@ impl AlgoSimple {
     }
 
     /// Get corresponding EccType by KeyType for Curve25519
-    fn get_ecc_type_25519(key_type: KeyType) -> EccType {
+    fn ecc_type_25519(key_type: KeyType) -> EccType {
         match key_type {
             KeyType::Signing
             | KeyType::Authentication
@@ -78,7 +73,7 @@ impl AlgoSimple {
     }
 
     /// Get corresponding Curve by KeyType for 25519 (Ed25519 vs Cv25519)
-    fn get_curve_for_25519(key_type: KeyType) -> Curve {
+    fn curve_for_25519(key_type: KeyType) -> Curve {
         match key_type {
             KeyType::Signing
             | KeyType::Authentication
@@ -87,49 +82,57 @@ impl AlgoSimple {
         }
     }
 
-    pub(crate) fn get_algo(&self, key_type: KeyType) -> Algo {
-        match self {
-            Self::RSA1k(e) => Algo::Rsa(RsaAttrs {
-                len_n: 1024,
-                len_e: *e,
-                import_format: 0,
-            }),
-            Self::RSA2k(e) => Algo::Rsa(RsaAttrs {
-                len_n: 2048,
-                len_e: *e,
-                import_format: 0,
-            }),
-            Self::RSA3k(e) => Algo::Rsa(RsaAttrs {
-                len_n: 3072,
-                len_e: *e,
-                import_format: 0,
-            }),
-            Self::RSA4k(e) => Algo::Rsa(RsaAttrs {
-                len_n: 4096,
-                len_e: *e,
-                import_format: 0,
-            }),
-            Self::NIST256 => Algo::Ecc(EccAttrs {
-                curve: Curve::NistP256r1,
-                ecc_type: Self::get_ecc_type(key_type),
-                import_format: None,
-            }),
-            Self::NIST384 => Algo::Ecc(EccAttrs {
-                curve: Curve::NistP384r1,
-                ecc_type: Self::get_ecc_type(key_type),
-                import_format: None,
-            }),
-            Self::NIST521 => Algo::Ecc(EccAttrs {
-                curve: Curve::NistP521r1,
-                ecc_type: Self::get_ecc_type(key_type),
-                import_format: None,
-            }),
-            Self::Curve25519 => Algo::Ecc(EccAttrs {
-                curve: Self::get_curve_for_25519(key_type),
-                ecc_type: Self::get_ecc_type_25519(key_type),
-                import_format: None,
-            }),
-        }
+    /// Return the appropriate Algo for this AlgoSimple.
+    ///
+    /// This mapping differs between cards, based on `ard` and `algo_info`
+    /// (e.g. the exact Algo variant can have a different size for e, in RSA;
+    /// also, the import_format can differ).
+    pub(crate) fn determine_algo(
+        &self,
+        key_type: KeyType,
+        ard: &ApplicationRelatedData,
+        algo_info: Option<AlgoInfo>,
+    ) -> Result<Algo> {
+        let algo = match self {
+            Self::RSA1k => Algo::Rsa(keys::determine_rsa_attrs(
+                1024, key_type, ard, algo_info,
+            )?),
+            Self::RSA2k => Algo::Rsa(keys::determine_rsa_attrs(
+                2048, key_type, ard, algo_info,
+            )?),
+            Self::RSA3k => Algo::Rsa(keys::determine_rsa_attrs(
+                3072, key_type, ard, algo_info,
+            )?),
+            Self::RSA4k => Algo::Rsa(keys::determine_rsa_attrs(
+                4096, key_type, ard, algo_info,
+            )?),
+            Self::NIST256 => Algo::Ecc(keys::determine_ecc_attrs(
+                Curve::NistP256r1.oid(),
+                Self::ecc_type(key_type),
+                key_type,
+                algo_info,
+            )?),
+            Self::NIST384 => Algo::Ecc(keys::determine_ecc_attrs(
+                Curve::NistP384r1.oid(),
+                Self::ecc_type(key_type),
+                key_type,
+                algo_info,
+            )?),
+            Self::NIST521 => Algo::Ecc(keys::determine_ecc_attrs(
+                Curve::NistP521r1.oid(),
+                Self::ecc_type(key_type),
+                key_type,
+                algo_info,
+            )?),
+            Self::Curve25519 => Algo::Ecc(keys::determine_ecc_attrs(
+                Self::curve_for_25519(key_type).oid(),
+                Self::ecc_type_25519(key_type),
+                key_type,
+                algo_info,
+            )?),
+        };
+
+        Ok(algo)
     }
 }
 
@@ -161,10 +164,30 @@ impl fmt::Display for Algo {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match self {
             Self::Rsa(rsa) => {
-                write!(f, "RSA {}, {} ", rsa.len_n, rsa.len_e)
+                write!(
+                    f,
+                    "RSA {} [e {}{}]",
+                    rsa.len_n,
+                    rsa.len_e,
+                    if rsa.import_format != 0 {
+                        format!(", format {}", rsa.import_format)
+                    } else {
+                        "".to_string()
+                    }
+                )
             }
             Self::Ecc(ecc) => {
-                write!(f, "{:?} ({:?})", ecc.curve, ecc.ecc_type)
+                write!(
+                    f,
+                    "{:?} ({:?}){}",
+                    ecc.curve,
+                    ecc.ecc_type,
+                    if ecc.import_format == Some(0xff) {
+                        " with pub"
+                    } else {
+                        ""
+                    }
+                )
             }
             Self::Unknown(u) => {
                 write!(f, "Unknown: {:?}", u)
@@ -176,7 +199,7 @@ impl fmt::Display for Algo {
 impl Algo {
     /// Get a DO representation of the Algo, for setting algorithm
     /// attributes on the card.
-    pub(crate) fn get_data(&self) -> Result<Vec<u8>, Error> {
+    pub(crate) fn to_data_object(&self) -> Result<Vec<u8>, Error> {
         match self {
             Algo::Rsa(rsa) => Self::rsa_algo_attrs(rsa),
             Algo::Ecc(ecc) => Self::ecc_algo_attrs(ecc.oid(), ecc.ecc_type()),
@@ -196,15 +219,6 @@ impl Algo {
         algo_attributes.push(0x00);
         algo_attributes.push(algo_attrs.len_e() as u8);
 
-        // Import-Format of private key
-        // (This fn currently assumes import_format "00 = standard (e, p, q)")
-        if algo_attrs.import_format() != 0 {
-            return Err(anyhow!(
-                "Unexpected RSA input format (only 0 is supported)"
-            )
-            .into());
-        }
-
         algo_attributes.push(algo_attrs.import_format());
 
         Ok(algo_attributes)
@@ -291,8 +305,13 @@ impl EccAttrs {
     pub fn oid(&self) -> &[u8] {
         self.curve.oid()
     }
+
+    pub fn import_format(&self) -> Option<u8> {
+        self.import_format
+    }
 }
 
+/// Enum for naming ECC curves, and mapping them to/from their OIDs.
 #[derive(Debug, Clone, Copy, Eq, PartialEq)]
 #[non_exhaustive]
 pub enum Curve {
