@@ -3,13 +3,13 @@
 
 //! CardApp exposes functionality of the "OpenPGP card" application.
 
-use std::borrow::BorrowMut;
 use std::convert::TryFrom;
 use std::convert::TryInto;
 
 use anyhow::{anyhow, Result};
 
 use crate::algorithm::{Algo, AlgoInfo, AlgoSimple};
+use crate::apdu::response::RawResponse;
 use crate::apdu::{commands, response::Response};
 use crate::card_do::{
     ApplicationRelatedData, CardholderRelatedData, Fingerprint,
@@ -19,7 +19,9 @@ use crate::crypto_data::{
     CardUploadableKey, Cryptogram, Hash, PublicKeyMaterial,
 };
 use crate::tlv::{tag::Tag, value::Value, Tlv};
-use crate::{apdu, keys, CardCaps, CardClientBox, KeyType};
+use crate::{
+    apdu, keys, CardCaps, CardClient, CardClientBox, KeyType, SmartcardError,
+};
 use crate::{Error, StatusBytes};
 
 /// Low-level access to OpenPGP card functionality.
@@ -33,29 +35,34 @@ pub struct CardApp {
     card_client: CardClientBox,
 }
 
-impl From<CardClientBox> for CardApp {
-    fn from(card_client: CardClientBox) -> Self {
-        Self { card_client }
-    }
-}
-
-impl From<CardApp> for CardClientBox {
-    fn from(card_app: CardApp) -> CardClientBox {
-        card_app.card_client
-    }
-}
-
 impl CardApp {
-    /// Get the CardClient for this CardApp
-    pub(crate) fn get_card_client(&mut self) -> &mut CardClientBox {
-        &mut self.card_client
+    /// Get a CardApp based on a CardClient.
+    ///
+    /// It is expected that SELECT has already been performed on the card
+    /// beforehand.
+    ///
+    /// This fn initializes the CardCaps by requesting
+    /// application_related_data from the card, and setting the
+    /// capabilities accordingly.
+    pub fn initialize(card_client: CardClientBox) -> Result<Self> {
+        let mut ca = Self { card_client };
+
+        let ard = ca.application_related_data()?;
+        ca.init_caps(&ard)?;
+
+        Ok(ca)
+    }
+
+    /// Get the CardClient of this CardApp
+    pub(crate) fn card_client(&mut self) -> &mut dyn CardClient {
+        &mut *self.card_client
     }
 
     /// Initialize the CardCaps settings in the underlying CardClient
     /// from the data in `ard`.
     ///
     /// This should be done at an early point, soon after opening the card.
-    pub fn init_caps(&mut self, ard: &ApplicationRelatedData) -> Result<()> {
+    fn init_caps(&mut self, ard: &ApplicationRelatedData) -> Result<()> {
         // Determine chaining/extended length support from card
         // metadata and cache this information in CardApp (as a
         // CardCaps)
@@ -63,41 +70,48 @@ impl CardApp {
         let mut ext_support = false;
         let mut chaining_support = false;
 
-        if let Ok(hist) = ard.get_historical() {
-            if let Some(cc) = hist.get_card_capabilities() {
-                chaining_support = cc.get_command_chaining();
-                ext_support = cc.get_extended_lc_le();
+        if let Ok(hist) = ard.historical_bytes() {
+            if let Some(cc) = hist.card_capabilities() {
+                chaining_support = cc.command_chaining();
+                ext_support = cc.extended_lc_le();
             }
         }
 
+        let ext_cap = ard.extended_capabilities()?;
+
+        // Get max command/response byte sizes from card
         let (max_cmd_bytes, max_rsp_bytes) =
-            if let Ok(Some(eli)) = ard.get_extended_length_information() {
+            if let Ok(Some(eli)) = ard.extended_length_information() {
+                // In card 3.x, max lengths come from ExtendedLengthInfo
                 (eli.max_command_bytes(), eli.max_response_bytes())
+            } else if let (Some(cmd), Some(rsp)) =
+                (ext_cap.max_cmd_len(), ext_cap.max_resp_len())
+            {
+                // In card 2.x, max lengths come from ExtendedCapabilities
+                (cmd, rsp)
             } else {
+                // Fallback: use 255 if we have no information from the card
                 (255, 255)
             };
 
+        let pw_status = ard.pw_status_bytes()?;
+        let pw1_max = pw_status.pw1_max_len();
+        let pw3_max = pw_status.pw3_max_len();
+
         let caps = CardCaps {
             ext_support,
             chaining_support,
             max_cmd_bytes,
             max_rsp_bytes,
+            pw1_max_len: pw1_max,
+            pw3_max_len: pw3_max,
         };
 
-        self.card_client.init_caps(caps);
+        self.card_client.init_card_caps(caps);
 
         Ok(())
     }
 
-    // --- select ---
-
-    /// Select the OpenPGP card application
-    pub fn select(&mut self) -> Result<Response, Error> {
-        let select_openpgp = commands::select_openpgp();
-        apdu::send_command(&mut self.card_client, select_openpgp, false)?
-            .try_into()
-    }
-
     // --- get data ---
 
     /// Get the "application related data" from the card.
@@ -105,11 +119,11 @@ impl CardApp {
     /// (This data should probably be cached in a higher layer. Some parts of
     /// it are needed regularly, and it does not usually change during
     /// normal use of a card.)
-    pub fn get_application_related_data(
+    pub fn application_related_data(
         &mut self,
     ) -> Result<ApplicationRelatedData> {
-        let ad = commands::get_application_data();
-        let resp = apdu::send_command(&mut self.card_client, ad, true)?;
+        let ad = commands::application_related_data();
+        let resp = apdu::send_command(self.card_client(), ad, true)?;
         let value = Value::from(resp.data()?, true)?;
 
         log::debug!(" App data Value: {:x?}", value);
@@ -117,75 +131,63 @@ impl CardApp {
         Ok(ApplicationRelatedData(Tlv::new(Tag::from([0x6E]), value)))
     }
 
-    /// Get data from "private use" DO.
-    ///
-    /// `num` must be between 1 and 4.
-    pub fn get_private(&mut self, num: u8) -> Result<Vec<u8>> {
-        assert!((1..=4).contains(&num));
-
-        let cmd = commands::get_private_do(num);
-        let resp = apdu::send_command(&mut self.card_client, cmd, true)?;
-
-        Ok(resp.data()?.to_vec())
-    }
-
-    fn get_ca_fingerprints() {
+    #[allow(dead_code)]
+    fn ca_fingerprints() {
         unimplemented!()
     }
 
-    fn get_key_information() {
+    #[allow(dead_code)]
+    fn key_information() {
         unimplemented!()
     }
 
-    fn get_uif_pso_cds() {
+    #[allow(dead_code)]
+    fn uif_pso_cds() {
         unimplemented!()
     }
 
-    fn get_uif_pso_dec() {
+    #[allow(dead_code)]
+    fn uif_pso_dec() {
         unimplemented!()
     }
 
-    fn get_uif_pso_aut() {
+    #[allow(dead_code)]
+    fn uif_pso_aut() {
         unimplemented!()
     }
 
-    fn get_uif_attestation() {
+    #[allow(dead_code)]
+    fn uif_attestation() {
         unimplemented!()
     }
 
-    // --- optional private DOs (0101 - 0104) ---
-
     // --- login data (5e) ---
 
-    // --- URL (5f50) ---
-
-    pub fn get_url(&mut self) -> Result<String> {
-        let resp = apdu::send_command(
-            &mut self.card_client,
-            commands::get_url(),
-            true,
-        )?;
+    /// Get URL (5f50)
+    pub fn url(&mut self) -> Result<String> {
+        let resp =
+            apdu::send_command(self.card_client(), commands::url(), true)?;
 
         Ok(String::from_utf8_lossy(resp.data()?).to_string())
     }
 
-    // --- cardholder related data (65) ---
-    pub fn get_cardholder_related_data(
+    /// Get cardholder related data (65)
+    pub fn cardholder_related_data(
         &mut self,
     ) -> Result<CardholderRelatedData> {
         let crd = commands::cardholder_related_data();
-        let resp = apdu::send_command(&mut self.card_client, crd, true)?;
+        let resp = apdu::send_command(self.card_client(), crd, true)?;
         resp.check_ok()?;
 
         CardholderRelatedData::try_from(resp.data()?)
     }
 
-    // --- security support template (7a) ---
-    pub fn get_security_support_template(
+    /// Get security support template (7a)
+    pub fn security_support_template(
         &mut self,
     ) -> Result<SecuritySupportTemplate> {
-        let sst = commands::get_security_support_template();
-        let resp = apdu::send_command(&mut self.card_client, sst, true)?;
+        let sst = commands::security_support_template();
+        let resp = apdu::send_command(self.card_client(), sst, true)?;
         resp.check_ok()?;
 
         let tlv = Tlv::try_from(resp.data()?)?;
@@ -211,16 +213,16 @@ impl CardApp {
     ///
     /// Call select_data() before calling this fn, to select a particular
     /// certificate (if the card supports multiple certificates).
-    pub fn get_cardholder_certificate(&mut self) -> Result<Response, Error> {
-        let cmd = commands::get_cardholder_certificate();
-        apdu::send_command(&mut self.card_client, cmd, true)?.try_into()
+    pub fn cardholder_certificate(&mut self) -> Result<Response, Error> {
+        let cmd = commands::cardholder_certificate();
+        apdu::send_command(self.card_client(), cmd, true)?.try_into()
     }
 
-    /// DO "Algorithm Information"
-    pub fn get_algo_info(&mut self) -> Result<Option<AlgoInfo>> {
+    /// Get "Algorithm Information"
+    pub fn algorithm_information(&mut self) -> Result<Option<AlgoInfo>> {
         let resp = apdu::send_command(
-            &mut self.card_client,
-            commands::get_algo_list(),
+            self.card_client(),
+            commands::algo_info(),
             true,
         )?;
         resp.check_ok()?;
@@ -229,8 +231,39 @@ impl CardApp {
         Ok(Some(ai))
     }
 
-    /// SELECT DATA "select a DO in the current template"
-    /// (e.g. for cardholder certificate)
+    /// Firmware Version (YubiKey specific (?))
+    pub fn firmware_version(&mut self) -> Result<Vec<u8>> {
+        let resp = apdu::send_command(
+            self.card_client(),
+            commands::firmware_version(),
+            true,
+        )?;
+
+        Ok(resp.data()?.into())
+    }
+
+    /// Set identity (Nitrokey Start specific (?)).
+    /// [see:
+    /// <https://docs.nitrokey.com/start/linux/multiple-identities.html>
+    /// <https://github.com/Nitrokey/nitrokey-start-firmware/pull/33/>]
+    pub fn set_identity(&mut self, id: u8) -> Result<Vec<u8>> {
+        let resp = apdu::send_command(
+            self.card_client(),
+            commands::set_identity(id),
+            false,
+        );
+
+        // Apparently it's normal to get "NotTransacted" from pcsclite when
+        // the identity switch was successful.
+        if let Err(Error::Smartcard(SmartcardError::NotTransacted)) = resp {
+            Ok(vec![])
+        } else {
+            Ok(resp?.data()?.into())
+        }
+    }
+
+    /// SELECT DATA ("select a DO in the current template",
+    /// e.g. for cardholder certificate)
     pub fn select_data(
         &mut self,
         num: u8,
@@ -244,7 +277,41 @@ impl CardApp {
         let data = tlv.serialize();
 
         let cmd = commands::select_data(num, data);
-        apdu::send_command(&mut self.card_client, cmd, true)?.try_into()
+        apdu::send_command(self.card_client(), cmd, true)?.try_into()
+    }
+
+    // --- optional private DOs (0101 - 0104) ---
+
+    /// Get data from "private use" DO.
+    ///
+    /// `num` must be between 1 and 4.
+    pub fn private_use_do(&mut self, num: u8) -> Result<Vec<u8>> {
+        assert!((1..=4).contains(&num));
+
+        let cmd = commands::private_use_do(num);
+        let resp = apdu::send_command(self.card_client(), cmd, true)?;
+
+        Ok(resp.data()?.to_vec())
+    }
+
+    /// Set data of "private use" DO.
+    ///
+    /// `num` must be between 1 and 4.
+    ///
+    /// Access condition:
+    /// - 1/3 need PW1 (82)
+    /// - 2/4 need PW3
+    pub fn set_private_use_do(
+        &mut self,
+        num: u8,
+        data: Vec<u8>,
+    ) -> Result<Vec<u8>> {
+        assert!((1..=4).contains(&num));
+
+        let cmd = commands::put_private_use_do(num, data);
+        let resp = apdu::send_command(self.card_client(), cmd, true)?;
+
+        Ok(resp.data()?.to_vec())
     }
 
     // ----------
@@ -252,20 +319,27 @@ impl CardApp {
     /// Reset all state on this OpenPGP card.
     ///
     /// Note: the "factory reset" operation is not directly offered by the
-    /// card. It is implemented as a series of OpenPGP card commands:
-    /// - send 4 bad requests to verify pw1
-    /// - send 4 bad requests to verify pw3
-    /// - terminate_df
-    /// - activate_file
+    /// card spec. It is implemented as a series of OpenPGP card commands:
+    /// - send 4 bad requests to verify pw1,
+    /// - send 4 bad requests to verify pw3,
+    /// - terminate_df,
+    /// - activate_file.
+    ///
+    /// With most cards, this sequence of operations causes the card
+    /// to revert to a "blank" state.
+    ///
+    /// (However, e.g. vanilla Gnuk doesn't support this functionality.
+    /// Gnuk needs to be built with the `--enable-factory-reset`
+    /// option to the `configure` script to enable this functionality).
     pub fn factory_reset(&mut self) -> Result<()> {
         // send 4 bad requests to verify pw1
         // [apdu 00 20 00 81 08 40 40 40 40 40 40 40 40]
         for _ in 0..4 {
             let verify = commands::verify_pw1_81([0x40; 8].to_vec());
-            let resp =
-                apdu::send_command(&mut self.card_client, verify, false)?;
+            let resp = apdu::send_command(self.card_client(), verify, false)?;
             if !(resp.status() == StatusBytes::SecurityStatusNotSatisfied
-                || resp.status() == StatusBytes::AuthenticationMethodBlocked)
+                || resp.status() == StatusBytes::AuthenticationMethodBlocked
+                || matches!(resp.status(), StatusBytes::PasswordNotChecked(_)))
             {
                 return Err(anyhow!("Unexpected status for reset, at pw1."));
             }
@@ -275,11 +349,11 @@ impl CardApp {
         // [apdu 00 20 00 83 08 40 40 40 40 40 40 40 40]
         for _ in 0..4 {
             let verify = commands::verify_pw3([0x40; 8].to_vec());
-            let resp =
-                apdu::send_command(&mut self.card_client, verify, false)?;
+            let resp = apdu::send_command(self.card_client(), verify, false)?;
 
             if !(resp.status() == StatusBytes::SecurityStatusNotSatisfied
-                || resp.status() == StatusBytes::AuthenticationMethodBlocked)
+                || resp.status() == StatusBytes::AuthenticationMethodBlocked
+                || matches!(resp.status(), StatusBytes::PasswordNotChecked(_)))
             {
                 return Err(anyhow!("Unexpected status for reset, at pw3."));
             }
@@ -287,19 +361,30 @@ impl CardApp {
 
         // terminate_df [apdu 00 e6 00 00]
         let term = commands::terminate_df();
-        let resp = apdu::send_command(&mut self.card_client, term, false)?;
+        let resp = apdu::send_command(self.card_client(), term, false)?;
         resp.check_ok()?;
 
         // activate_file [apdu 00 44 00 00]
         let act = commands::activate_file();
-        let resp = apdu::send_command(&mut self.card_client, act, false)?;
+        let resp = apdu::send_command(self.card_client(), act, false)?;
         resp.check_ok()?;
 
         Ok(())
     }
 
-    /// Verify pw1 (user) for signing operation (mode 81) and set an
-    /// appropriate access status.
+    // --- verify/modify ---
+
+    /// Does the cardreader support direct pinpad verify?
+    pub fn feature_pinpad_verify(&self) -> bool {
+        self.card_client.feature_pinpad_verify()
+    }
+
+    /// Does the cardreader support direct pinpad modify?
+    pub fn feature_pinpad_modify(&self) -> bool {
+        self.card_client.feature_pinpad_modify()
+    }
+
+    /// Verify pw1 (user) for signing operation (mode 81).
     ///
     /// Depending on the PW1 status byte (see Extended Capabilities) this
     /// access condition is only valid for one PSO:CDS command or remains
@@ -309,7 +394,21 @@ impl CardApp {
         pin: &str,
     ) -> Result<Response, Error> {
         let verify = commands::verify_pw1_81(pin.as_bytes().to_vec());
-        apdu::send_command(&mut self.card_client, verify, false)?.try_into()
+        apdu::send_command(self.card_client(), verify, false)?.try_into()
+    }
+
+    /// Verify pw1 (user) for signing operation (mode 81) using a
+    /// pinpad on the card reader. If no usable pinpad is found, an error
+    /// is returned.
+    ///
+    /// Depending on the PW1 status byte (see Extended Capabilities) this
+    /// access condition is only valid for one PSO:CDS command or remains
+    /// valid for several attempts.
+    pub fn verify_pw1_for_signing_pinpad(
+        &mut self,
+    ) -> Result<Response, Error> {
+        let res = self.card_client.pinpad_verify(0x81)?;
+        RawResponse::try_from(res)?.try_into()
     }
 
     /// Check the current access of PW1 for signing (mode 81).
@@ -317,17 +416,26 @@ impl CardApp {
     /// If verification is not required, an empty Ok Response is returned.
     ///
     /// (Note: some cards don't correctly implement this feature,
-    /// e.g. yubikey 5)
+    /// e.g. YubiKey 5)
     pub fn check_pw1_for_signing(&mut self) -> Result<Response, Error> {
         let verify = commands::verify_pw1_81(vec![]);
-        apdu::send_command(&mut self.card_client, verify, false)?.try_into()
+        apdu::send_command(self.card_client(), verify, false)?.try_into()
     }
 
-    /// Verify PW1 (user) and set an appropriate access status.
+    /// Verify PW1 (user).
     /// (For operations except signing, mode 82).
     pub fn verify_pw1(&mut self, pin: &str) -> Result<Response, Error> {
         let verify = commands::verify_pw1_82(pin.as_bytes().to_vec());
-        apdu::send_command(&mut self.card_client, verify, false)?.try_into()
+        apdu::send_command(self.card_client(), verify, false)?.try_into()
+    }
+
+    /// Verify PW1 (user) for operations except signing (mode 82),
+    /// using a pinpad on the card reader. If no usable pinpad is found,
+    /// an error is returned.
+
+    pub fn verify_pw1_pinpad(&mut self) -> Result<Response, Error> {
+        let res = self.card_client.pinpad_verify(0x82)?;
+        RawResponse::try_from(res)?.try_into()
     }
 
     /// Check the current access of PW1.
@@ -336,16 +444,23 @@ impl CardApp {
     /// If verification is not required, an empty Ok Response is returned.
     ///
     /// (Note: some cards don't correctly implement this feature,
-    /// e.g. yubikey 5)
+    /// e.g. YubiKey 5)
     pub fn check_pw1(&mut self) -> Result<Response, Error> {
         let verify = commands::verify_pw1_82(vec![]);
-        apdu::send_command(&mut self.card_client, verify, false)?.try_into()
+        apdu::send_command(self.card_client(), verify, false)?.try_into()
     }
 
-    /// Verify PW3 (admin) and set an appropriate access status.
+    /// Verify PW3 (admin).
     pub fn verify_pw3(&mut self, pin: &str) -> Result<Response, Error> {
         let verify = commands::verify_pw3(pin.as_bytes().to_vec());
-        apdu::send_command(&mut self.card_client, verify, false)?.try_into()
+        apdu::send_command(self.card_client(), verify, false)?.try_into()
+    }
+
+    /// Verify PW3 (admin) using a pinpad on the card reader. If no usable
+    /// pinpad is found, an error is returned.
+    pub fn verify_pw3_pinpad(&mut self) -> Result<Response, Error> {
+        let res = self.card_client.pinpad_verify(0x83)?;
+        RawResponse::try_from(res)?.try_into()
     }
 
     /// Check the current access of PW3 (admin).
@@ -353,10 +468,10 @@ impl CardApp {
     /// If verification is not required, an empty Ok Response is returned.
     ///
     /// (Note: some cards don't correctly implement this feature,
-    /// e.g. yubikey 5)
+    /// e.g. YubiKey 5)
     pub fn check_pw3(&mut self) -> Result<Response, Error> {
         let verify = commands::verify_pw3(vec![]);
-        apdu::send_command(&mut self.card_client, verify, false)?.try_into()
+        apdu::send_command(self.card_client(), verify, false)?.try_into()
     }
 
     /// Change the value of PW1 (user password).
@@ -372,7 +487,14 @@ impl CardApp {
         data.extend(new.as_bytes());
 
         let change = commands::change_pw1(data);
-        apdu::send_command(&mut self.card_client, change, false)?.try_into()
+        apdu::send_command(self.card_client(), change, false)?.try_into()
+    }
+
+    /// Change the value of PW1 (user password)  using a pinpad on the
+    /// card reader. If no usable pinpad is found, an error is returned.
+    pub fn change_pw1_pinpad(&mut self) -> Result<Response, Error> {
+        let res = self.card_client.pinpad_modify(0x81)?;
+        RawResponse::try_from(res)?.try_into()
     }
 
     /// Change the value of PW3 (admin password).
@@ -388,7 +510,14 @@ impl CardApp {
         data.extend(new.as_bytes());
 
         let change = commands::change_pw3(data);
-        apdu::send_command(&mut self.card_client, change, false)?.try_into()
+        apdu::send_command(self.card_client(), change, false)?.try_into()
+    }
+
+    /// Change the value of PW3 (admin password) using a pinpad on the
+    /// card reader. If no usable pinpad is found, an error is returned.
+    pub fn change_pw3_pinpad(&mut self) -> Result<Response, Error> {
+        let res = self.card_client.pinpad_modify(0x83)?;
+        RawResponse::try_from(res)?.try_into()
     }
 
     /// Reset the error counter for PW1 (user password) and set a new value
@@ -404,7 +533,7 @@ impl CardApp {
         resetting_code: Option<Vec<u8>>,
     ) -> Result<Response, Error> {
         let reset = commands::reset_retry_counter_pw1(resetting_code, new_pw1);
-        apdu::send_command(&mut self.card_client, reset, false)?.try_into()
+        apdu::send_command(self.card_client(), reset, false)?.try_into()
     }
 
     // --- decrypt ---
@@ -416,16 +545,20 @@ impl CardApp {
     pub fn decipher(&mut self, dm: Cryptogram) -> Result<Vec<u8>, Error> {
         match dm {
             Cryptogram::RSA(message) => {
+                // "Padding indicator byte (00) for RSA" (pg. 69)
                 let mut data = vec![0x0];
-
-                // FIXME: The spec says we should "format according to PKCS#1"
-
                 data.extend_from_slice(message);
 
                 // Call the card to decrypt `data`
                 self.pso_decipher(data)
             }
             Cryptogram::ECDH(eph) => {
+                // "In case of ECDH the card supports a partial decrypt
+                // only. The input is a cipher DO with the following data:"
+                // A6 xx Cipher DO
+                //  -> 7F49 xx Public Key DO
+                //    -> 86 xx External Public Key
+
                 // External Public Key
                 let epk = Tlv::new([0x86], Value::S(eph.to_vec()));
 
@@ -445,7 +578,7 @@ impl CardApp {
     fn pso_decipher(&mut self, data: Vec<u8>) -> Result<Vec<u8>, Error> {
         // The OpenPGP card is already connected and PW1 82 has been verified
         let dec_cmd = commands::decryption(data);
-        let resp = apdu::send_command(&mut self.card_client, dec_cmd, true)?;
+        let resp = apdu::send_command(self.card_client(), dec_cmd, true)?;
         resp.check_ok()?;
 
         Ok(resp.data().map(|d| d.to_vec())?)
@@ -453,16 +586,8 @@ impl CardApp {
 
     // --- sign ---
 
-    /// Sign `hash`, on the card.
-    ///
-    /// (This is a wrapper around the low-level
-    /// pso_compute_digital_signature operation.
-    /// It builds the required `data` field from `hash`)
-    pub fn signature_for_hash(
-        &mut self,
-        hash: Hash,
-    ) -> Result<Vec<u8>, Error> {
-        let data = match hash {
+    fn digestinfo(hash: Hash) -> Vec<u8> {
+        match hash {
             Hash::SHA256(_) | Hash::SHA384(_) | Hash::SHA512(_) => {
                 let tlv = Tlv::new(
                     [0x30],
@@ -486,62 +611,91 @@ impl CardApp {
             }
             Hash::EdDSA(d) => d.to_vec(),
             Hash::ECDSA(d) => d.to_vec(),
-        };
+        }
+    }
 
-        self.pso_compute_digital_signature(data)
+    /// Sign `hash`, on the card.
+    ///
+    /// This is a wrapper around the low-level
+    /// pso_compute_digital_signature operation.
+    /// It builds the required `data` field from `hash`.
+    ///
+    /// For RSA, this means a "DigestInfo" data structure is generated.
+    /// (see 7.2.10.2 DigestInfo for RSA).
+    ///
+    /// With ECC the hash data is processed as is, using
+    /// pso_compute_digital_signature.
+    pub fn signature_for_hash(
+        &mut self,
+        hash: Hash,
+    ) -> Result<Vec<u8>, Error> {
+        self.pso_compute_digital_signature(Self::digestinfo(hash))
     }
 
     /// Run signing operation on the smartcard (low level operation)
     /// (7.2.10 PSO: COMPUTE DIGITAL SIGNATURE)
-    fn pso_compute_digital_signature(
+    pub fn pso_compute_digital_signature(
         &mut self,
         data: Vec<u8>,
     ) -> Result<Vec<u8>, Error> {
-        let dec_cmd = commands::signature(data);
+        let cds_cmd = commands::signature(data);
 
-        let resp = apdu::send_command(&mut self.card_client, dec_cmd, true)?;
+        let resp = apdu::send_command(self.card_client(), cds_cmd, true)?;
+
+        Ok(resp.data().map(|d| d.to_vec())?)
+    }
+
+    // --- internal authenticate ---
+
+    /// Auth-sign `hash`, on the card.
+    ///
+    /// This is a wrapper around the low-level
+    /// internal_authenticate operation.
+    /// It builds the required `data` field from `hash`.
+    ///
+    /// For RSA, this means a "DigestInfo" data structure is generated.
+    /// (see 7.2.10.2 DigestInfo for RSA).
+    ///
+    /// With ECC the hash data is processed as is.
+    pub fn authenticate_for_hash(
+        &mut self,
+        hash: Hash,
+    ) -> Result<Vec<u8>, Error> {
+        self.internal_authenticate(Self::digestinfo(hash))
+    }
+
+    /// Run signing operation on the smartcard (low level operation)
+    /// (7.2.13 INTERNAL AUTHENTICATE)
+    pub fn internal_authenticate(
+        &mut self,
+        data: Vec<u8>,
+    ) -> Result<Vec<u8>, Error> {
+        let ia_cmd = commands::internal_authenticate(data);
+        let resp = apdu::send_command(self.card_client(), ia_cmd, true)?;
 
         Ok(resp.data().map(|d| d.to_vec())?)
     }
 
     // --- admin ---
 
-    /// Set data of "private use" DO.
-    ///
-    /// `num` must be between 1 and 4.
-    ///
-    /// Access condition:
-    /// - 1/3 need PW1 (82)
-    /// - 2/4 need PW3
-    pub fn set_private(&mut self, num: u8, data: Vec<u8>) -> Result<Vec<u8>> {
-        assert!((1..=4).contains(&num));
-
-        let cmd = commands::put_private_do(num, data);
-        let resp = apdu::send_command(&mut self.card_client, cmd, true)?;
-
-        Ok(resp.data()?.to_vec())
-    }
-
     pub fn set_name(&mut self, name: &str) -> Result<Response, Error> {
         let put_name = commands::put_name(name.as_bytes().to_vec());
-        apdu::send_command(&mut self.card_client, put_name, false)?.try_into()
+        apdu::send_command(self.card_client(), put_name, false)?.try_into()
     }
 
     pub fn set_lang(&mut self, lang: &str) -> Result<Response, Error> {
         let put_lang = commands::put_lang(lang.as_bytes().to_vec());
-        apdu::send_command(self.card_client.borrow_mut(), put_lang, false)?
-            .try_into()
+        apdu::send_command(self.card_client(), put_lang, false)?.try_into()
     }
 
     pub fn set_sex(&mut self, sex: Sex) -> Result<Response, Error> {
         let put_sex = commands::put_sex((&sex).into());
-        apdu::send_command(self.card_client.borrow_mut(), put_sex, false)?
-            .try_into()
+        apdu::send_command(self.card_client(), put_sex, false)?.try_into()
     }
 
     pub fn set_url(&mut self, url: &str) -> Result<Response, Error> {
         let put_url = commands::put_url(url.as_bytes().to_vec());
-        apdu::send_command(&mut self.card_client, put_url, false)?.try_into()
+        apdu::send_command(self.card_client(), put_url, false)?.try_into()
     }
 
     pub fn set_creation_time(
@@ -558,12 +712,10 @@ impl CardApp {
             .copied()
             .collect();
 
-        let time_cmd = commands::put_data(
-            &[key_type.get_timestamp_put_tag()],
-            time_value,
-        );
+        let time_cmd =
+            commands::put_data(&[key_type.timestamp_put_tag()], time_value);
 
-        apdu::send_command(&mut self.card_client, time_cmd, false)?.try_into()
+        apdu::send_command(self.card_client(), time_cmd, false)?.try_into()
     }
 
     pub fn set_fingerprint(
@@ -572,11 +724,11 @@ impl CardApp {
         key_type: KeyType,
     ) -> Result<Response, Error> {
         let fp_cmd = commands::put_data(
-            &[key_type.get_fingerprint_put_tag()],
+            &[key_type.fingerprint_put_tag()],
             fp.as_bytes().to_vec(),
         );
 
-        apdu::send_command(self.get_card_client(), fp_cmd, false)?.try_into()
+        apdu::send_command(self.card_client(), fp_cmd, false)?.try_into()
     }
 
     /// Set PW Status Bytes.
@@ -598,7 +750,7 @@ impl CardApp {
         let data = pw_status.serialize_for_put(long);
 
         let cmd = commands::put_pw_status(data);
-        apdu::send_command(self.get_card_client(), cmd, false)?.try_into()
+        apdu::send_command(self.card_client(), cmd, false)?.try_into()
     }
 
     /// Set cardholder certificate (for AUT, DEC or SIG).
@@ -610,7 +762,7 @@ impl CardApp {
         data: Vec<u8>,
     ) -> Result<Response, Error> {
         let cmd = commands::put_cardholder_certificate(data);
-        apdu::send_command(&mut self.card_client, cmd, false)?.try_into()
+        apdu::send_command(self.card_client(), cmd, false)?.try_into()
     }
 
     /// Set algorithm attributes
@@ -622,11 +774,11 @@ impl CardApp {
     ) -> Result<Response, Error> {
         // Command to PUT the algorithm attributes
         let cmd = commands::put_data(
-            &[key_type.get_algorithm_tag()],
-            algo.get_data()?,
+            &[key_type.algorithm_tag()],
+            algo.to_data_object()?,
         );
 
-        apdu::send_command(&mut self.card_client, cmd, false)?.try_into()
+        apdu::send_command(self.card_client(), cmd, false)?.try_into()
     }
 
     /// Set resetting code
@@ -636,7 +788,7 @@ impl CardApp {
         resetting_code: Vec<u8>,
     ) -> Result<Response, Error> {
         let cmd = commands::put_data(&[0xd3], resetting_code);
-        apdu::send_command(&mut self.card_client, cmd, false)?.try_into()
+        apdu::send_command(self.card_client(), cmd, false)?.try_into()
     }
 
     /// Import an existing private key to the card.
@@ -646,13 +798,13 @@ impl CardApp {
         key: Box<dyn CardUploadableKey>,
         key_type: KeyType,
     ) -> Result<(), Error> {
-        let algo_list = self.get_algo_info();
+        let algo_info = self.algorithm_information();
 
         // An error is ok - it's fine if a card doesn't offer a list of
         // supported algorithms
-        let algo_list = algo_list.unwrap_or(None);
+        let algo_info = algo_info.unwrap_or(None);
 
-        keys::key_import(self, key, key_type, algo_list)
+        keys::key_import(self, key, key_type, algo_info)
     }
 
     /// Generate a key on the card.
@@ -660,6 +812,10 @@ impl CardApp {
     ///
     /// If the `algo` parameter is Some, then this algorithm will be set on
     /// the card for "key_type".
+    ///
+    /// Note: `algo` needs to precisely specify the RSA bitsize of e (if
+    /// applicable), and import format, with values that the current card
+    /// supports.
     pub fn generate_key(
         &mut self,
         fp_from_pub: fn(
@@ -676,8 +832,12 @@ impl CardApp {
     /// Generate a key on the card.
     /// (7.2.14 GENERATE ASYMMETRIC KEY PAIR)
     ///
-    /// This is a convenience wrapper around generate_key() which allows
+    /// This is a wrapper around generate_key() which allows
     /// using the simplified `AlgoSimple` algorithm selector enum.
+    ///
+    /// Note: AlgoSimple doesn't specify card specific details (such as
+    /// bitsize of e for RSA, and import format). This function determines
+    /// these values based on information from the card.
     pub fn generate_key_simple(
         &mut self,
         fp_from_pub: fn(
@@ -686,22 +846,28 @@ impl CardApp {
             KeyType,
         ) -> Result<Fingerprint, Error>,
         key_type: KeyType,
-        algo: AlgoSimple,
+        simple: AlgoSimple,
     ) -> Result<(PublicKeyMaterial, KeyGenerationTime), Error> {
-        let algo = algo.get_algo(key_type);
+        let ard = self.application_related_data()?;
+        let algo_info = self.algorithm_information()?;
+
+        let algo = simple.determine_algo(key_type, &ard, algo_info)?;
+
         self.generate_key(fp_from_pub, key_type, Some(&algo))
     }
 
     /// Get public key material from the card.
     ///
-    /// Note: this fn returns an uninterpreted set of raw values (not an
-    /// OpenPGP key data structure).
-    /// This data from the card is insufficient to create a typical
-    /// full public key.
-    pub fn get_pub_key(
+    /// Note: this fn returns a set of raw public key data (not an
+    /// OpenPGP data structure).
+    ///
+    /// Note also that the information from the card is insufficient to
+    /// reconstruct a pre-existing OpenPGP public key that corresponds to
+    /// the private key on the card.
+    pub fn public_key(
         &mut self,
         key_type: KeyType,
     ) -> Result<PublicKeyMaterial, Error> {
-        keys::get_pub_key(self, key_type)
+        keys::public_key(self, key_type)
     }
 }
