@@ -18,10 +18,10 @@ impl PWStatusBytes {
     pub(crate) fn serialize_for_put(&self, long: bool) -> Vec<u8> {
         let mut data = vec![];
 
-        data.push(if !self.pw1_cds_multi { 0 } else { 1 });
+        data.push(if self.pw1_cds_valid_once { 0 } else { 1 });
 
         if long {
-            let mut b2 = self.pw1_len;
+            let mut b2 = self.pw1_len_format;
             if self.pw1_pin_block {
                 b2 |= 0x80;
             }
@@ -29,7 +29,7 @@ impl PWStatusBytes {
 
             data.push(self.rc_len);
 
-            let mut b4 = self.pw3_len;
+            let mut b4 = self.pw3_len_format;
             if self.pw3_pin_block {
                 b4 |= 0x80;
             }
@@ -45,7 +45,7 @@ impl TryFrom<&[u8]> for PWStatusBytes {
 
     fn try_from(input: &[u8]) -> Result<Self, Self::Error> {
         if input.len() == 7 {
-            let pw1_cds_multi = input[0] == 0x01;
+            let pw1_cds_valid_once = input[0] == 0x00;
             let pw1_pin_block = input[1] & 0x80 != 0;
             let pw1_len = input[1] & 0x7f;
             let rc_len = input[2];
@@ -56,12 +56,12 @@ impl TryFrom<&[u8]> for PWStatusBytes {
             let err_count_pw3 = input[6];
 
             Ok(Self {
-                pw1_cds_multi,
+                pw1_cds_valid_once,
                 pw1_pin_block,
-                pw1_len,
+                pw1_len_format: pw1_len,
                 rc_len,
                 pw3_pin_block,
-                pw3_len,
+                pw3_len_format: pw3_len,
                 err_count_pw1,
                 err_count_rst,
                 err_count_pw3,
@@ -90,12 +90,12 @@ mod test {
         assert_eq!(
             pws,
             PWStatusBytes {
-                pw1_cds_multi: false,
+                pw1_cds_valid_once: true,
                 pw1_pin_block: false,
-                pw1_len: 0x40,
+                pw1_len_format: 0x40,
                 rc_len: 0x40,
                 pw3_pin_block: false,
-                pw3_len: 0x40,
+                pw3_len_format: 0x40,
                 err_count_pw1: 3,
                 err_count_rst: 0,
                 err_count_pw3: 3
