@@ -22,6 +22,8 @@ pub enum Hash<'a> {
 }
 
 impl Hash<'_> {
+    /// This fn is currently only used in the context of creating a
+    /// digestinfo for SHA*. Other OIDs are not implemented.
     pub(crate) fn oid(&self) -> Option<&'static [u8]> {
         match self {
             Self::SHA256(_) => {
@@ -33,8 +35,8 @@ impl Hash<'_> {
             Self::SHA512(_) => {
                 Some(&[0x60, 0x86, 0x48, 0x01, 0x65, 0x03, 0x04, 0x02, 0x03])
             }
-            Self::EdDSA(_) => panic!("This should not be called"),
-            Self::ECDSA(_) => panic!("This should not be called"),
+            Self::EdDSA(_) => panic!("OIDs for EdDSA are unimplemented"),
+            Self::ECDSA(_) => panic!("OIDs for ECDSA are unimplemented"),
         }
     }
 
@@ -65,13 +67,13 @@ pub enum Cryptogram<'a> {
 /// to an OpenPGP card
 pub trait CardUploadableKey {
     /// private key data
-    fn get_key(&self) -> Result<PrivateKeyMaterial>;
+    fn private_key(&self) -> Result<PrivateKeyMaterial>;
 
     /// timestamp of (sub)key creation
-    fn get_ts(&self) -> KeyGenerationTime;
+    fn timestamp(&self) -> KeyGenerationTime;
 
     /// fingerprint
-    fn get_fp(&self) -> Result<Fingerprint, Error>;
+    fn fingerprint(&self) -> Result<Fingerprint, Error>;
 }
 
 /// Algorithm-independent container for private key material to upload to
@@ -85,18 +87,30 @@ pub enum PrivateKeyMaterial {
 /// RSA-specific container for private key material to upload to an OpenPGP
 /// card.
 pub trait RSAKey {
-    fn get_e(&self) -> &[u8];
-    fn get_n(&self) -> &[u8];
-    fn get_p(&self) -> &[u8];
-    fn get_q(&self) -> &[u8];
+    // FIXME: use a mechanism like sequoia_openpgp::crypto::mem::Protected
+    // for private key material?
+
+    fn e(&self) -> &[u8];
+    fn p(&self) -> &[u8];
+    fn q(&self) -> &[u8];
+
+    fn pq(&self) -> Box<[u8]>;
+    fn dp1(&self) -> Box<[u8]>;
+    fn dq1(&self) -> Box<[u8]>;
+
+    fn n(&self) -> &[u8];
 }
 
 /// ECC-specific container for private key material to upload to an OpenPGP
 /// card.
 pub trait EccKey {
-    fn get_oid(&self) -> &[u8];
-    fn get_scalar(&self) -> &[u8];
-    fn get_type(&self) -> EccType;
+    // FIXME: use a mechanism like sequoia_openpgp::crypto::mem::Protected
+    // for private key material?
+
+    fn oid(&self) -> &[u8];
+    fn private(&self) -> Vec<u8>;
+    fn public(&self) -> Vec<u8>;
+    fn ecc_type(&self) -> EccType;
 }
 
 /// Algorithm-independent container for public key material retrieved from
@@ -108,6 +122,26 @@ pub enum PublicKeyMaterial {
     E(EccPub),
 }
 
+impl std::fmt::Display for PublicKeyMaterial {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        use hex_slice::AsHex;
+
+        match self {
+            Self::R(rsa) => {
+                write!(
+                    f,
+                    "RSA, n: {:02X}, e: {:02X}",
+                    rsa.n.plain_hex(false),
+                    rsa.v.plain_hex(false)
+                )
+            }
+            Self::E(ecc) => {
+                write!(f, "ECC, data: {:02X}", ecc.data.plain_hex(false))
+            }
+        }
+    }
+}
+
 /// RSA-specific container for public key material from an OpenPGP card.
 #[derive(Debug)]
 #[non_exhaustive]
